% Copyright (c) 2011-2012, Tom Van Cutsem, Vrije Universiteit Brussel
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%    * Neither the name of the Vrije Universiteit Brussel nor the
%      names of its contributors may be used to endorse or promote products
%      derived from this software without specific prior written permission.
%
%THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
%ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
%WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
%DISCLAIMED. IN NO EVENT SHALL VRIJE UNIVERSITEIT BRUSSEL BE LIABLE FOR ANY
%DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
%(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES
%LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
%ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
%(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
%SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

% A toy instant messaging program, used to illustrate simple concurrent
% programming with processes in Erlang

% To test this code, first make sure you define a file ~/.erlang.cookie
% with a random string in it.
% Next, start two OS shells and cd to the directory in which
% this file is contained. Then, in the first shell, type:
% erl -sname server@localhost
% > c(chat).
% > chat:start_chat_server().
% 
% In the second shell, type:
% erl -sname client@localhost
% > c(chat).
% > C1 = chat:start_chat_client("nickname").
% > C1 ! logon.
%
% The client should now communicate with the server, and
% should be successfully logged in. The client can send a
% message to itself:
% > C1 ! {send_message, "nickname", "Hi"}.

-module (chat).
-export ([chat_server/0, chat_client/1, start_chat_server/0, start_chat_client/1]).

-define(SERVER_NAME, chat_server).
-define(SERVER_NODE, server@localhost).
-define(CLIENT_TIMEOUT, 5000).

chat_server() ->
  process_flag(trap_exit, true),
  chat_server([]).

% Clients = [ { ClientPid, Nick } ]  
chat_server(Clients) ->
  receive
    {logon, ClientPid, Nick} ->
      NewClients = srv_logon(ClientPid,Nick,Clients),
      chat_server(NewClients);
    {logoff, ClientPid} ->
      NewClients = srv_logoff(ClientPid,Clients),
      chat_server(NewClients);
    {route_message, FromClientPid, ToNick, Text} ->
      srv_route(FromClientPid, ToNick, Text, Clients),
      chat_server(Clients);
    {'EXIT', ClientPid, Reason} -> % client exit = automatic logoff
      io:format("Client terminated, auto log off: ~p~n", [Reason]),
      NewClients = srv_logoff(ClientPid, Clients),
      chat_server(NewClients)
  end.

srv_logon(ClientPid, Nick, Clients) ->
  case lists:keymember(ClientPid, 1, Clients) of
    true ->
      io:format("Client ~p already logged in~n", [Nick]),
      ClientPid ! {error, already_logged_in},
      Clients;
    false ->
      io:format("Client ~p logged in~n", [Nick]),
      link(ClientPid), % if Client process crashes, this server needs to know
      ClientPid ! ok,
      [{ClientPid, Nick} | Clients]
  end.
  
srv_logoff(ClientPid, Clients) ->
  case lists:keysearch(ClientPid, 1, Clients) of
    {value, {_, Nick}} ->
      io:format("Client ~p logged off~n", [Nick]),
      ClientPid ! ok,
      lists:keydelete(ClientPid, 1, Clients);
    false ->
      io:format("Client not logged in: ~p~n", [ClientPid]),
      ClientPid ! {error, not_logged_in},
      Clients
  end.
  
srv_route(FromClientPid, ToNick, Text, Clients) ->
  case lists:keysearch(FromClientPid, 1, Clients) of
    {value, {_, FromNick}} ->
      srv_route(FromClientPid, FromNick, ToNick, Text, Clients);
    false ->
      io:format("Client not logged in:~p~n", [FromClientPid]),
      FromClientPid ! {error, not_logged_in}
  end.
srv_route(FromClientPid, FromNick, ToNick, Text, Clients) ->
  case lists:keysearch(ToNick, 2, Clients) of
    {value, {ToClientPid, _}} ->
      io:format("Routing message from ~p to ~p~n", [FromNick,ToNick]),
      ToClientPid ! {receive_message, FromNick, Text},
      FromClientPid ! ok;
    false ->
      io:format("Recipient ~p not logged in~n", [ToNick]),
      FromClientPid ! {error, recipient_not_logged_in}
  end.
  
start_chat_server() ->
  register(?SERVER_NAME, spawn(chat, chat_server, [])).

chat_client(Nick) ->
  receive
    logon ->
      io:format("[~p] Trying to log in...~n", [Nick]),
      {?SERVER_NAME, ?SERVER_NODE} ! {logon, self(), Nick},
      client_await_answer(Nick),
      chat_client(Nick);
    logoff ->
      io:format("[~p] Trying to log off...~n", [Nick]),
      {?SERVER_NAME, ?SERVER_NODE} ! {logoff, self()},
      client_await_answer(Nick),
      chat_client(Nick);
    {send_message, ToNick, Text} ->
      io:format("[~p] Trying to send message to ~p~n", [Nick, ToNick]),
      {?SERVER_NAME, ?SERVER_NODE} ! {route_message, self(), ToNick, Text},
      client_await_answer(Nick),
      chat_client(Nick);
    {receive_message, FromNick, Text} ->
      io:format("[~p] Incoming message from ~p: ~p~n", [Nick, FromNick, Text]),
      chat_client(Nick)
  end.
  
client_await_answer(Nick) ->
  receive
    ok ->
      io:format("[~p] ok~n",[Nick]);
    {error, Reason} ->
      io:format("[~p] failed: ~p~n", [Nick, Reason])
  after ?CLIENT_TIMEOUT ->
      io:format("[~p] request timed out~n", [Nick])
  end.

start_chat_client(Nick) ->
  spawn(chat, chat_client, [Nick]).